.tagtask {
    margin: 5px 0;
}

.tag, .task {
    border: 1px solid #888;
    border-radius: 5px;
    padding: 3px;
}
.tag {
    background: rgb(252, 224, 242);
}
.task {
    background: rgb(213, 250, 250);
}

.dataset-id {
    font-weight: bold;
    font-size: x-large;
    border-left: 3px solid black;
    padding-left: 5px;
}

.dataset-name {
    font-style: italic;
}

.dataset-entry {
    margin: 40px 0px;
}

.dataset-entry .description {
    color: #444
}
